from .modelling import Model

class Pipeline(Model):
    def __init__(self, model, featureslice):
        self.model = model
        self.featureslice = featureslice
        self.__name__ = self.model.__name__ + ':' + "undefined_fs"

    def fit(self, **kwargs):
        tmp = self.featureslice()
        self.model.fit(tmp.values, self.featureslice.target.values, **kwargs)

    def predict(self, df, **kwargs):
        return self.model.predict(self.featureslice(df).values, **kwargs)


from . import cache_utils
from .. import config
from glob import glob
import os


class Info:
    def __init__(self):
        super().__setattr__('__attributes', dict())
        self.recalc()

    def recalc(self):
        for name in glob(config.info_path + '*'):
            super().__getattribute__('__attributes')[name.split('/')[-1][:-5]] = cache_utils.load_obj(name)
            
    def __setattr__(self, key, value):
        super().__getattribute__('__attributes')[key] = value
        cache_utils.save_obj(value, cache_utils.get_path_info(key))

    def __contains__(self, item):
        return item in super().__getattribute__('__attributes')

    def __getattr__(self, key):
        self.recalc()
        if key in super().__getattribute__('__attributes'):
            return super().__getattribute__('__attributes')[key]
        else:
            raise KeyError

    def __delattr__(self, key):
        raise AttributeError("info is read-only")
            
    def __del__(self):
        for key, value in self.__attributes.items():
            print(key, value)
            cache_utils.save_obj(value, cache_utils.get_path_info(key))


info = Info()



import numpy as np

from .. import config
from .experiment import experiment_list, Experiment
from ..pipeline import Pipeline
# from ..feature.storage import FeatureSlice
from copy import deepcopy
import tqdm

class Validator:
    def __init__(self, splitter, metric, enable_widget=False):
        self.splitter = splitter
        self.metric = metric
        self.bar = (tqdm.tqdm_notebook if enable_widget else tqdm.tqdm)
        
    def score(self, model, featureset, **fit_params):
        pipelines = []
        scores = []
        oofs = np.zeros_like(featureset.target, dtype=np.float)
        weights = np.zeros_like(featureset.target, dtype=np.float)
        pbar = self.bar(self.splitter.split, total=self.splitter.size)
        pbar.set_description_str(f"Val of {model.__name__}")
        for spl in pbar:
            idx_train = spl['train']
            idx_test = spl['test']
            c_model = deepcopy(model)
            pl = Pipeline(c_model, featureset.slice(idx_train))
            pl.fit(**fit_params)
            pred = pl.predict(idx_test)
            pl.featureslice.compress()
            oofs[idx_test] = (weights[idx_test] * oofs[idx_test] + pred) / (weights[idx_test] + 1)
            weights[idx_test] += 1
            # print(featureset.target[idx_test].values[:10], pred[:10])
            score = self.metric(featureset.target.values[idx_test], pred)
            pipelines.append(pl)
            scores.append(score)
            pbar.set_postfix_str(f"score: {np.mean(scores)}")
        model_name = f"ens_{pipelines[0].model.__name__}_x{len(pipelines)}"
        final_ensemble = pipelines[0]
        for pipeline in pipelines[1:]:
            final_ensemble = final_ensemble + pipeline
        final_ensemble = final_ensemble / len(pipelines)
        final_ensemble.__name__ = model_name
        score = np.mean(scores)
        std = np.std(scores)
        experiment_list.register(
            Experiment(
                final_ensemble,
                oofs,
                score,
                std
            )
        )
        return score


from ..modelling import *
from .. import config

from xgboost import XGBClassifier as _XGBC
class XGBClassifier(Model):
    Estimator = _XGBC
    tracked_params = ['n_estimators', 'max_depth']
    fit_params = {'verbose': False}
    short_name = 'xgb'
    
from lightgbm import LGBMClassifier as _LGBMC
class LGBMClassifier(Model):
    Estimator = _LGBMC
    search_spaces = {
        'num_leaves': (2, 50),
        'learning_rate': (0.01, 1.0),
        'max_depth': (0, 50),
        'min_data_in_leaf': (20, 200),
        'max_bin': (2, 255),
    }

    system_params = {
        "metric": 'binary_logloss',
        "verbosity": 1,
        "seed": config.seed}
    short_name = 'lgb'
    
    def predict(self, X):
        return self.estimator.predict_proba(X)[:, 1]
    
from catboost import CatBoostClassifier as _CBC
class CBClassifier(Model):
    Estimator = _CBC
    short_name = 'cb'
    def __init__(self, params=None, n_jobs=-1, verbosity=0):
        super().__init__(params, n_jobs, verbosity)
        self.verbose = verbosity
        self.estimator.thread_count = n_jobs
        # self.estimator.verbose = verbosity

    def _fit(self, X, y, **kwargs):
        self.estimator.fit(X, y, verbose=self.verbose)

    def predict(self, X):
        return self.estimator.predict_proba(X)[:, 1]

from sklearn.linear_model import LogisticRegression as _LOGR
class LogisticRegression(Model):
    Estimator = _LOGR
    short_name = 'logr'
    search_spaces = {
        'C': [10**x for x in range(-4, 5)],
        'penalty': ['l1', 'l2'],
        'fit_intercept': [True, False],
        'class_weight': [None, 'balanced']
    }
    # max_iter and warm_start are useless for liblinear solver
    system_params = {
        'tol': 0.0001,
        'solver': 'liblinear',
        'multi_class': 'auto',
        'n_jobs': -1,
        'verbose': 0,
        'random_state': 42
    }

    def predict(self, X):
        return self.estimator.predict_proba(X)[:, 1]
